create or replace package callgraph_pac
  authid definer 
is
/*******************************************************************************
  <B>Callgraph</B> 
  
  Werkzeuge um Callgraphen von Paketen (und Funktionen/Prozeduren) zu erstellen.

  Mit Hilfe dieses Pakets knnen von allen dem Anwender sichtbaren, ffentlichen
  Funktionen/Prozeduren von Paketen Ausgaben erzeugt werden, mit deren Hilfe
  dot/graphviz Callgraphen erzeugen kann.
  http://www.graphviz.org/ 

  Dank an Duane Dieterich fr den Blog-Artikel
  http://www.reldesgen.com/2015/06/plsql-call-trees-from-dbaidentifiers.html
  mit dessen Hilfe dieses Paket erstellt wurde.

  Das Paket kann fr jeden/s beliebige/n User/Schema/Projekt eingesetzt werden.
  
  Um das Paket nutzen zu knnen, muss im verwendeten Schema eine Tabelle
  
  create table  fkt_call
  (
    --
    --  caller/Aufrufer
    --
    callr_owner             varchar2(128) null,
    callr_object_name       varchar2(128) null,
    callr_object_name_orig  varchar2(128) null,
    callr_name              varchar2(128) null,
    callr_signature         varchar2(32)  null,
    callr_usage             varchar2(11)  null,
    callr_usage_id          number        null,
    callr_usage_context_id  number        null,
    --
    --  callee/Aufgeruferer
    --
    callee_owner              varchar2(128) null,
    callee_object_name        varchar2(128) null,
    callee_object_name_orig   varchar2(128) null,    
    callee_name               varchar2(128) null,
    callee_signature          varchar2(32)  null,
    callee_usage              varchar2(11)  null,
    callee_usage_id           number        null,
    callee_usage_context_id   number        null
  );
  
  vorhanden sein. 
  Der verwendete User muss das Recht haben delete/insert/update auf diese
  Tabelle auszufhren.
  Die auszuwertenden Objekte/Pakete mssen mit   
  
  ALTER SESSION SET PLSCOPE_SETTINGS='IDENTIFIERS:ALL';
  oder 
  ALTER SESSION SET PLSCOPE_SETTINGS='IDENTIFIERS:PLSQL';  (ab Oracle 12.2)
  
  compiliert sein (d.h. die Tabelle sys.all_identifiers (public.all_identifiers)
  muss mit den notwendigen Daten befllt sein).
  
  Verwendung:
  
  get_all():  Mit dieser Prozedur werden alle verfgbaren Aufrufe ermittelt 
              und aufbereitet in der Tabelle function_calls abgelegt.
              Wichtiger Hinweise: 
              - Diese Prozedur muss fr jedes auszuwertende Schema einmal
                ausgefhrt werden (Falls sich die Struktur der Aufrufe in 
                den Paketen ndert muss es wiederholt werden).
              - Die Ausfhrung dieser Prozedur kann mehrere Minuten dauern!!
                    
  get_graph(...):   Erzeugt fr eine angegebene (ffentliche) Funktion/Prozedur
                    eine Ausgabe, die als Eingabe fr dot/graphviz verwendet 
                    werden kann. 
                    Die Ausgabe muss in eine Datei geschrieben werden, die
                    als Eingabeparameter von dot verwendet wird.
                    berladen: 
                    - Funktion: Ausgabe als clob
                    - Prozedur: Ausgabe per dbms_output

  %Created_Date   01.2021
  %Created_By     PL DAME-INA (OFD KA)
  %Object_Type    PACKAGE
  %Object_Name    callgraph_pac
  %Object_Owner   
  %Comments       -

  %Author         F0290597/j.demuth (jd)

  %change 2021-01-25  1.0 (jd)  erstellt
  %change 2021-03-09  2.0 (jd)  komplett berarbeitet
*******************************************************************************/


/*******************************************************************************
  Ermittelt mit Hilfe des Public Synonym all_identifiers alle Funktionsaufrufe
  die der angemeldete Benutzer sehen kann.
  Bereitet diese in der Tabelle fkt_call auf, so dass diese fr die 
  Verwendung mit get_graph(...) zur Verfgung stehen.
  
  Hinweise:
  - Einige Funktionsaufrufe knnen (im Moment) noch nicht aufgelst werden!
    Z.B. solche fkt(), die sich innerhalb von dbms_output.put_line( fkt() );
    befinden.
    
  - Es knnten auch andere Identifier (Tabellen, Typen, Cursor...) 
    oder SQL-Statements aufbereitet und anschlieend visualisiert werden.

  %author J. Demuth (jd)

  %change 2021-01-26  (jd)  erstellt
  %change 2021-03-09  (jd)  komplett berarbeitet
*******************************************************************************/
procedure get_all;


/*******************************************************************************

  Erzeugt fr die in den Parametern angegebene Funktion/Prozedur den Quelltext
  fr dot um daraus die Graphik eines Callgraphen erstellen zu knnen.
  
  berladen: 
  - Die Funktion gibt den erzeugten dot-Quelltext als clob zurck.
  - Die Prozedur gibt den erzeugten dot-Quelltext mittels dbms_output aus.
    (...und kann in sqlplus mittels spool in eine Datei geschrieben werden)

  %param    p_owner         Das Schema/der Owner der gewnschten Funktion
  %param    p_object_name   Der Name des Objekts (Paketname, 
                            bei Einzelfunktionen/-paketen deren Name)
  %param    p_name          Der Name der Funktion der Prozedur deren Callgraph 
                            erzeugt werden soll.

  %return   Der erzeugte dot-Quelltext

  %author   J. Demuth (jd)

  %change   2021-02-02  (jd)  erstellt
  %chnge    2021-03-08  (jd)  komplett berarbeitet
*******************************************************************************/
function get_graph( p_owner       in varchar2,
                    p_object_name in varchar2,
                    p_name        in varchar2 )
  return clob;

procedure get_graph( p_owner       in varchar2,
                     p_object_name in varchar2,
                     p_name        in varchar2 );

end callgraph_pac;
/
create or replace package body callgraph_pac is
/*******************************************************************************
  Siehe Package

  %Created_Date   01.2021
  %Created_By     PL DAME-INA (OFD KA)
  %Object_Type    PACKAGE
  %Object_Name    callgraph_pac
  %Object_Owner
  %Comments       -

  %Author         F0290597/j.demuth (jd)

  %change 2021-01-25  1.0 (jd)  erstellt
  %change 2021-03-09  2.0 (jd)  komplett berarbeitet
*******************************************************************************/

LF            constant VARCHAR2(1)  := chr(10);
CR            constant VARCHAR2(1)  := chr(13);

-- um LF,CR, ... einfach umschalten zu knnen ((endline)).
endl          constant varchar2(2) := CR || LF;

/*******************************************************************************
* Der Cursor fr die Ertellung der rekursiven Aufrufe kann ...
* ... durch individuelle Einschrnkungen angepasst werden.
* ... durch upper() an Klein/Gro-Schreibung angepasst werden.
*******************************************************************************/
cursor cur_graph( in_owner        in varchar2,
                  in_object_name  in varchar2,
                  in_name         in varchar2 )
  is
    select distinct
      case f.callr_owner when in_owner then null else f.callr_owner || '.' end
        || case when f.callr_owner       = in_owner 
                 and f.callr_object_name_orig = in_object_name then null 
                else f.callr_object_name_orig || '.' end
        || f.callr_name                                           as callr,
      case f.callee_owner when in_owner then null else f.callee_owner || '.' end
        || case when f.callee_owner = in_owner 
                 and f.callee_object_name_orig = in_object_name then null 
                else f.callee_object_name_orig || '.' end
        || f.callee_name                                          as callee,
      min(level)                                                  as level_min,
      CONNECT_BY_ISCYCLE                                          as is_cycle,
      f.callr_owner,  f.callr_object_name_orig,  f.callr_name,
      f.callee_owner, f.callee_object_name_orig, f.callee_name
    from fkt_call f
    where f.callee_owner != 'SYS'
      -- individuelle Einschrnkungen hier
      and f.callr_object_name  != 'LOG_PAC'
      and f.callee_object_name != 'LOG_PAC'
      --
      and not regexp_like( f.callr_name,  '^LOG( \(\d\))?$' ) -- berladungen
      and not regexp_like( f.callee_name, '^LOG( \(\d\))?$' )
      -- auf eigene Funktionen beschrnken
      and f.callr_owner       = in_owner
      and f.callr_object_name = in_object_name
    connect by nocycle 
                       prior f.callee_owner       = f.callr_owner
                   and prior f.callee_object_name = f.callr_object_name
                   and prior f.callee_name        = f.callr_name
    start with f.callr_owner       = in_owner
           and f.callr_object_name = in_object_name
           and regexp_like( f.callr_name, '^' || in_name || '( \(\d+\))?$' ) -- berladungen
    group by
      f.callr_owner,  f.callr_object_name_orig,  f.callr_name,
      f.callee_owner, f.callee_object_name_orig, f.callee_name,
      CONNECT_BY_ISCYCLE
    order by
      min(level),
      f.callr_owner,  f.callr_object_name_orig,  f.callr_name,
      f.callee_owner, f.callee_object_name_orig, f.callee_name;

-- Fr rekursive selbst-Aufrufe (indirekte werden bereits von cur_graph behandelt)
cursor cur_graph_rec( in_owner        in varchar2,
                      in_object_name  in varchar2,
                      in_name         in varchar2 )
  is with calls as 
    ( select distinct
        min(level)                                                  as level_min,
        CONNECT_BY_ISCYCLE                                          as is_cycle,
        f.callr_owner,  f.callr_object_name_orig,  f.callr_name,
        f.callee_owner, f.callee_object_name_orig, f.callee_name
      from fkt_call f
      where f.callee_owner != 'SYS'
        -- individuelle Einschrnkungen hier
      and f.callr_object_name  != 'LOG_PAC'
      and f.callee_object_name != 'LOG_PAC'
      --
      and not regexp_like( f.callr_name,  '^LOG( \(\d\))?$' ) -- berladungen
      and not regexp_like( f.callee_name, '^LOG( \(\d\))?$' )
        and 'LOG_IT' not in ( f.callr_name, f.callee_name )         
        -- auf eigene Funktionen beschrnken
        and f.callr_owner       = in_owner
        and f.callr_object_name = in_object_name
      connect by nocycle 
                         prior f.callee_owner       = f.callr_owner
                     and prior f.callee_object_name = f.callr_object_name
                     and prior f.callee_name        = f.callr_name
      start with f.callr_owner       = in_owner
             and f.callr_object_name = in_object_name
             and regexp_like( f.callr_name, '^' || in_name || '( \(\d+\))?$' ) -- berladungen
      group by
        f.callr_owner,  f.callr_object_name_orig,  f.callr_name,
        f.callee_owner, f.callee_object_name_orig, f.callee_name,
        CONNECT_BY_ISCYCLE 
    )
  select distinct
        case f.callr_owner when in_owner then null else f.callr_owner || '.' end
          || case when f.callr_owner       = in_owner 
                   and f.callr_object_name_orig = in_object_name then null 
                  else f.callr_object_name_orig || '.' end
          || f.callr_name                                           as callr,
        case f.callee_owner when in_owner then null else f.callee_owner || '.' end
          || case when f.callee_owner = in_owner 
                   and f.callee_object_name_orig = in_object_name then null 
                  else f.callee_object_name_orig || '.' end
          || f.callee_name                                          as callee
  from 
    fkt_call f, 
    calls c 
  where c.callee_owner             = f.callee_owner 
    and c.callee_object_name_orig  = f.callee_object_name_orig 
    and c.callee_name              = f.callee_name      
    --
    and f.callr_owner             = f.callee_owner 
    and f.callr_object_name_orig  = f.callee_object_name_orig 
    and f.callr_name              = f.callee_name;

-- Vorwrtsdeklaration
procedure get_callees;
procedure get_caller;
procedure mark_overload;


/*******************************************************************************
********************************************************************************
* Public
********************************************************************************
*******************************************************************************/


/*******************************************************************************
* Siehe Schnittstelle
*******************************************************************************/
procedure get_all
is
  pragma autonomous_transaction;
begin
  get_callees;
  get_caller; 
  mark_overload;

  commit;  
end get_all;


/*******************************************************************************
* Siehe Schnittstelle
*******************************************************************************/
function get_graph( p_owner       in varchar2,
                    p_object_name in varchar2,
                    p_name        in varchar2 )
  return clob
is

  anz binary_integer := 0;
  ret clob;
begin

  ret := 'digraph plsql_callgraph'  || endl
      || '{'                        || endl
      || '  rankdir=LR;'            || endl
      || '  node [shape=box];'      || endl;

  for rec in cur_graph( p_owner, p_object_name, p_name ) loop
    ret := ret || '"' || lower(rec.callr)  || '" -> "'
                      || lower(rec.callee) || '"' || endl;
    anz := anz + 1;
  end loop;

  for rec in cur_graph_rec( p_owner, p_object_name, p_name ) loop
    ret := ret || '  "' || lower(rec.callr)  || '" -> "'
                        || lower(rec.callee) || '"' || endl;
    anz := anz + 1;
  end loop;

  if 0 = anz then
    ret := ret || '  "' || lower(p_name) || '"' || endl;
  end if;

  ret := ret || '}';

  return ret;
end get_graph;


/*******************************************************************************
* Siehe Schnittstelle
*******************************************************************************/
procedure get_graph( p_owner       in varchar2,
                     p_object_name in varchar2,
                     p_name        in varchar2 )
is
  
  anz binary_integer := 0;
begin
  dbms_output.enable();

  dbms_output.put_line( 'digraph plsql_callgraph' );
  dbms_output.put_line( '{' );
  dbms_output.put_line( '  rankdir=LR;' );
  dbms_output.put_line( '  node [shape=box];' );

  for rec in cur_graph( p_owner, p_object_name, p_name ) loop
    dbms_output.put_line( '  "' || lower(rec.callr)  || '" -> "'
                                || lower(rec.callee) || '"' );
    anz := anz + 1;
  end loop;

  for rec in cur_graph_rec( p_owner, p_object_name, p_name ) loop
    dbms_output.put_line( '  "' || lower(rec.callr)  || '" -> "'
                                || lower(rec.callee) || '"' );
    anz := anz + 1;
  end loop;

  if 0 = anz then
    dbms_output.put_line( '  "' || lower(p_name) || '"' );
  end if;

  dbms_output.put_line( '}' );

end get_graph;


/*******************************************************************************
********************************************************************************
* Private
********************************************************************************
*******************************************************************************/


/*******************************************************************************

  Ermittelt alle Aufrufe, die der angemeldete Benutzer sehen kann und
  trgt diese in fkt_call ein.

  Hinweise: 
  - Die Aufrufer werden hier noch nicht erittelt.
  - An dieser Stelle knnen Einschrnkungen bzgl. der spter anzuzeigenden
    Aufrufe implementiert werden: 
    + Typ des Aufrufs
    + Besitzer des aufgerufenen Objekts
    + ...
    Standardmig werden Prozeduren und Funktionen angezeigt, 
    die nicht SYS oder UT3 gehren. (UT3 http://utplsql.org)

  %author   J. Demuth (jd)

  %change   2021-03-08  (jd)  erstellt
*******************************************************************************/
procedure get_callees
is
begin
  delete fkt_call;

  insert into fkt_call
    ( callee_owner,
      callee_object_name,
      callee_name,
      callee_signature,
      callee_usage,
      callee_usage_id,
      callee_usage_context_id )
  select
    i.OWNER,
    i.OBJECT_NAME,
    i.NAME,
    i.SIGNATURE,
    i.USAGE,
    i.USAGE_ID,
    i.USAGE_CONTEXT_ID
  from 
    all_identifiers i,
    all_identifiers i_sys
  where i.USAGE = 'CALL'
    and i.TYPE in ( 'PROCEDURE', 'FUNCTION', 'SYNONYM' ) -- 'NESTED TABLE', 'CURSOR', ...
    and i.SIGNATURE = i_sys.SIGNATURE
    and i_sys.USAGE = 'DECLARATION'
    and i_sys.OWNER not in ( 'SYS', 'UT3' ) ;    

end get_callees;


/*******************************************************************************

  Ermittelt rekursiv alle Aufrufer, die eine der zuvor ermittelten Funktionen 
  oder Prozeduren aufruft und trgt diese in fkt_call ein.
  
  Hinweis: Fr die Anzeige wird der originale .object_name ermittelt. 
           Dieser weicht vom .object_name in all_identifiers ab bei
           Funktionan/Prozeduren aus anderen Objekten (Paketen) als dem,
           in dem der Aufruf stattfindet.

  %author   J. Demuth (jd)

  %change   2021-03-08  (jd)  erstellt
*******************************************************************************/
procedure get_caller
is
  cursor cur_callee is
    select
      f.callee_owner,
      f.callee_object_name,
      f.callee_name,
      f.callee_signature,
      f.callee_usage_context_id
    from fkt_call f
    for update;

  cursor cur_callr( in_owner      in varchar2,
                    in_obj_name   in varchar2,
                    in_use_con_id in number )
  is
    select
      i.OWNER,
      i.OBJECT_NAME,
      i.NAME,
      i.SIGNATURE,
      i.TYPE,
      i.USAGE,
      i.USAGE_ID,
      i.USAGE_CONTEXT_ID
    from
      all_identifiers   i
    where i.owner       = in_owner
     and i.object_name  = in_obj_name
     and i.usage_id     = in_use_con_id;
     
  callr           cur_callr%rowtype;

  cursor cur_obj_repair( in_owner       in varchar2,
                         in_signature   in varchar2,
                         in_object_name in varchar2 )
  is
    select i.OBJECT_NAME
    from all_identifiers i
    where i.OWNER       = in_owner
      and i.SIGNATURE   = in_signature
      and i.OBJECT_NAME != in_object_name
      and i.USAGE       = 'DECLARATION';                            

  obj_repair        cur_obj_repair%rowtype;

  callr_repair        varchar2(128);  
  callee_repair       varchar2(128);

  usage_context_id   number;
  weiter             boolean;
  -- gegen Endlosschleife aus unbekannten Grnden
  anz                 binary_integer; 

begin
  for callee in cur_callee loop

    usage_context_id  := callee.callee_usage_context_id;
    weiter            := true;
    anz               := 0;
        
    -- maximale Rekursionstiefe: 1000
    while weiter and anz < 1000 loop
      callr := null;
      open cur_callr( callee.callee_owner, 
                      callee.callee_object_name, 
                      usage_context_id );
      fetch cur_callr into callr;
      close cur_callr;
      
      if    callr.usage_context_id is null 
        or  callr.usage_context_id = 0 
      then
        weiter := false;
      else
        usage_context_id := callr.usage_context_id;
      end if;
      
      if    callr.usage = 'DEFINITION' 
        and callr.type in ( 'FUNCTION', 'PROCEDURE' ) -- sonst endet es z.B. bei 'VARIABLE'
      then
        weiter := false;  
        
        obj_repair.object_name := null;
        open cur_obj_repair( callr.owner, 
                             callr.signature,
                             callr.object_name );
        fetch cur_obj_repair into obj_repair;
        close cur_obj_repair;
                
        callr_repair := coalesce( obj_repair.object_name, 
                                  callr.object_name );
        
        obj_repair.object_name := null;
        open cur_obj_repair( callee.callee_owner, 
                             callee.callee_signature,
                             callee.callee_object_name );
        fetch cur_obj_repair into obj_repair;
        close cur_obj_repair;
        
        callee_repair := coalesce( obj_repair.object_name, 
                                   callee.callee_object_name );
        
        update fkt_call f
        set f.callr_owner             = callr.owner,
            f.callr_object_name       = callr.object_name,
            f.callr_object_name_orig  = callr_repair,
            f.callr_name              = callr.name,
            f.callr_signature         = callr.signature,
            f.callr_usage             = callr.usage,
            f.callr_usage_id          = callr.usage_id,
            f.callr_usage_context_id  = callr.usage_context_id,
            f.callee_object_name_orig = callee_repair
        where current of cur_callee;
        
      end if; --  if callr.usage = 'DECLARATION' then
    
      anz := anz + 1;
    end loop; --  while weiter loop
    
  end loop; --  for callee in cur_callee loop
  
end get_caller;  


/*******************************************************************************

  Ermittelt berladene Funktionen/Prozeduren und markiert diese mit einer
  fortlaufenden Nummer "<name> (<laufende Nummer>)" 
  (z.B. test (1), test (2), ...)
  um sie in den Aufrufen unterscheiden zu knnen.

  %author   J. Demuth (jd)

  %change   2021-03-08  (jd)  erstellt
*******************************************************************************/
procedure mark_overload
is
  -- all_procedures.overload reicht nicht,
  -- da private berladungen da nicht enthalten sind!
  cursor cur_o is
    with overload as  ( select
                          i.OWNER,
                          i.OBJECT_NAME,
                          i.NAME,
                          count(*)
                        from
                          all_identifiers i
                        where i.USAGE = 'DECLARATION'
                          and i.TYPE in ( 'FUNCTION', 'PROCEDURE' )
                        group by i.OWNER, i.OBJECT_NAME, i.NAME
                        having count(*) > 1
                      )
      select
        i.SIGNATURE,
        i.NAME || ' (' || to_char( row_number() over ( partition by i.OWNER, i.OBJECT_NAME, i.NAME
                                                       order by i.SIGNATURE )
                                 ) || ')' as name_mod,
        i.name,
        i.OBJECT_NAME,
        i.OWNER
      from all_identifiers  i,
           overload         o
      where i.OWNER = o.owner
        and i.OBJECT_NAME = o.object_name
        and i.NAME = o.name
        and i.USAGE = 'DECLARATION'
        and i.TYPE in ( 'FUNCTION', 'PROCEDURE' )
        and i.OWNER not in ( 'SYS', 'UT3' ) ;
begin
  for rec_o in cur_o loop
    update fkt_call f
    set f.callr_name    = rec_o.name_mod
    where f.callr_owner             = rec_o.owner
      and f.callr_object_name_orig  = rec_o.object_name
      and f.callr_name              = rec_o.name
      and f.callr_signature         = rec_o.signature;

    update fkt_call f
    set f.callee_name   = rec_o.name_mod
    where f.callee_owner            = rec_o.owner
      and f.callee_object_name_orig = rec_o.object_name
      and f.callee_name             = rec_o.name
      and f.callee_signature        = rec_o.signature;
  end loop;
  
end mark_overload;


end callgraph_pac;
/
